#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#ifdef TURBOC
#include <dos.h>
#include <alloc.h>
#include <graphics.h>
#endif
#include "wirefram.h"

extern VIEWPOINT *Eye;
extern OBJECT *Light_Sources;
extern OBJECT *Current_Display_Object;
extern int XResolution;
extern int YResolution;
extern unsigned long First_Map_Line, Last_Map_Line;
extern void display_title_string(char *title);

#ifndef M_PI
#define M_PI 3.1415926535897932384626
#endif
#ifndef HUGE
#define HUGE 1.0e6
#endif
#ifndef EPSILON
#define EPSILON 1.0e-6
#endif

#define DEBUG_RENDERER

#define MakeVec(a, b, c, v)      (v).x=(a),(v).y=(b),(v).z=(c)
#define VecScale(S,a)   (a).x *= S ; (a).y *= S ; (a).z *= S
#define VecNegate(a)   (a).x=0.0-(a).x;(a).y=0.0-(a).y;(a).z=0.0-(a).z;
#define VecDot(a,b)   ((a).x*(b).x+(a).y*(b).y+(a).z*(b).z)
#define VecLen(a)   (sqrt(VecDot(a,a)))
#define VecCopy(a,b)    (b).x=(a).x;(b).y=(a).y;(b).z=(a).z;
#define VecAdd(a,b,c) (c).x=(a).x+(b).x;(c).y=(a).y+(b).y;(c).z=(a).z+(b).z
#define VecSub(a,b,c) (c).x=(a).x-(b).x;(c).y=(a).y-(b).y;(c).z=(a).z-(b).z
#define VecCross(a,b,c)    (c).x=(a).y*(b).z-(a).z*(b).y;\
          (c).y=(a).z*(b).x-(a).x*(b).z;\
          (c).z=(a).x*(b).y-(a).y*(b).x

static Poly_vert *poly_dummy;  /* used superficially by POLY_MASK macro */
static Poly_box box;
static Window win;
static DBL Phong_Coeff = 20.0;
static DBL Global_Ambient = 0.2;
static DBL Global_Diffuse = 0.8;

void poly_print(/* char *, p */);
void poly_vert_label(/* str, mask */);
void poly_vert_print(/* str, v, mask */);
int  poly_clip_to_box(Poly *, Poly_box *);
void poly_clip_to_halfspace(Poly *, Poly *, int, DBL, DBL);

void poly_scan(Poly *, Window *, void (*)());
void scan_convert(Poly *, TRANSFORMATION *);

/* Screen, zbuffer, and shading routines */
static void Determine_Color(VECTOR, VECTOR, VECTOR *);
static void shade_point(VECTOR, VECTOR, VECTOR, DBL, DBL *, DBL *);

/* Declarations of vector/matrix operations */
void VecH(VECTOR, VECTOR, VECTOR *);
DBL VecNormalize(VECTOR *);

/* Special object renderer routines */
void render(Poly *, void (*)(Poly *, DBL, DBL, Poly_vert *), TRANSFORMATION *);

static void incrementalize_y(DBL *, DBL *, DBL *, DBL *, int, int);
static void incrementalize_x(DBL *, DBL *, DBL *, DBL *, int, int);
static void scanline(int, Poly_vert *, Poly_vert *, Window *,
                     void (*)(int, int, Poly_vert *), int);
static void increment(DBL *, DBL *, int);

DBL 
VecNormalize(vec)
 VECTOR  *vec;
{
   DBL len, x2, y2, z2;
   x2 = vec->x * vec->x;
   y2 = vec->y * vec->y;
   z2 = vec->z * vec->z;
   len = x2 + y2 + z2;
   if (len < EPSILON) {
      printf("Bad vector: <%f, %f, %f> -> %f in VecNormalize\n",
             vec->x, vec->y, vec->z, len);
      printf("x^2 = %f, y^2 = %f, z^2 = %f\n", x2, y2, z2);
      exit(-1);
      }
   else {
      len = sqrt(len);
      VInverseScale(*vec, *vec, len);
      }
   return(len) ;
}

void
VecH(a, b, c)
   VECTOR a, b, *c;
{
   VAdd(*c, a, b);
   if (VecNormalize(c) < 0.0) {
      printf("Half vector is null\n");
      exit(1);
      }
}

#ifdef DEBUG_RENDERER
void poly_vert_label(char *, int);
void poly_vert_print(char *, Poly_vert *, int);

void
poly_vert_label(str, mask)
   char *str;
   int mask;
{
   printf("%s", str);
   if (mask&POLY_MASK(sx))   printf("   sx  ");
   if (mask&POLY_MASK(sy))   printf("   sy  ");
   if (mask&POLY_MASK(sz))   printf("   sz  ");
   if (mask&POLY_MASK(sw))   printf("   sw  ");
   if (mask&POLY_MASK(x))    printf("   x   ");
   if (mask&POLY_MASK(y))    printf("   y   ");
   if (mask&POLY_MASK(z))    printf("   z   ");
   if (mask&POLY_MASK(nx))   printf("   nx  ");
   if (mask&POLY_MASK(ny))   printf("   ny  ");
   if (mask&POLY_MASK(nz))   printf("   nz  ");
   printf("\n");
}

void
poly_vert_print(str, v, mask)
   char *str;
   Poly_vert *v;
   int mask;
{
   printf("%s", str);
   if (mask&POLY_MASK(sx)) printf(" %6.1f", v->sx);
   if (mask&POLY_MASK(sy)) printf(" %6.1f", v->sy);
   if (mask&POLY_MASK(sz)) printf(" %6.2f", v->sz);
   if (mask&POLY_MASK(sw)) printf(" %6.2f", v->sw);
   if (mask&POLY_MASK(x))  printf(" %6.2f", v->x);
   if (mask&POLY_MASK(y))  printf(" %6.2f", v->y);
   if (mask&POLY_MASK(z))  printf(" %6.2f", v->z);
   if (mask&POLY_MASK(nx)) printf(" %6.3f", v->nx);
   if (mask&POLY_MASK(ny)) printf(" %6.3f", v->ny);
   if (mask&POLY_MASK(nz)) printf(" %6.3f", v->nz);
   printf("\n");
}

/* poly_print: print Poly p to stdout, prefixed by the label str */
void
poly_print(str, p)
   char *str;
   Poly *p;
{
   int i;

   printf("%s: %d sides\n", str, p->n);
   poly_vert_label("        ", p->mask);
   for (i=0; i<p->n; i++) {
      printf("   v[%d] ", i);
      poly_vert_print("", &p->vert[i], p->mask);
      }
}
#endif

/*
 * poly_clip.c: homogeneous 3-D convex polygon clipper
 *
 * Paul Heckbert        1985, Dec 1989
 */

#define SWAP(a, b, temp)        {temp = a; a = b; b = temp;}
#define COORD(vert, i)  ((DBL *)(vert))[i]

#define CLIP_AND_SWAP(elem, sign, k, p, q, r) { \
    poly_clip_to_halfspace(p, q, &v->elem-(DBL *)v, sign, sign*k); \
    if (q->n==0) {p1->n = 0; return POLY_CLIP_OUT;} \
    SWAP(p, q, r); \
}

/*
 * poly_clip_to_box: Clip the convex polygon p1 to the screen space box
 * using the homogeneous screen coordinates (sx, sy, sz, sw) of each              
 * vertex, testing if v->sx/v->sw > box->x0 and v->sx/v->sw < box->x1, 
 * and similar tests for y and z, for each vertex v of the polygon.
 * If polygon is entirely inside box, then POLY_CLIP_IN is returned.
 * If polygon is entirely outside box, then POLY_CLIP_OUT is returned.
 * Otherwise, if the polygon is cut by the box, p1 is modified and
 * POLY_CLIP_PARTIAL is returned.
 */

int
poly_clip_to_box(p1, box)
   Poly *p1;
   Poly_box *box;
{
   int x0out = 0, x1out = 0;
   int y0out = 0, y1out = 0;
   int z0out = 0, z1out = 0;
   int i;
   Poly_vert *v;
   Poly p2, *p, *q, *r;

   /* count vertices "outside" with respect to each */
   /* of the six planes */
   for (v=p1->vert, i=p1->n; i>0; i--, v++) {
      if (v->sx < box->x0 * v->sw) x0out++;     /* out on left */
      if (v->sx > box->x1 * v->sw) x1out++;     /* out on right */
      if (v->sy < box->y0 * v->sw) y0out++;     /* out on top */
      if (v->sy > box->y1 * v->sw) y1out++;     /* out on bottom */
      if (v->sz < box->z0 * v->sw) z0out++;     /* out on near */
      if (v->sz > box->z1 * v->sw) z1out++;     /* out on far */
      }
   /* check if all vertices inside */
   if (x0out+x1out+y0out+y1out+z0out+z1out == 0)
      return POLY_CLIP_IN;

   /* check if all vertices are "outside" any of the six planes */
   if (x0out==p1->n || x1out==p1->n || y0out==p1->n ||
       y1out==p1->n || z0out==p1->n || z1out==p1->n) {
      p1->n = 0;
      return POLY_CLIP_OUT;
      }
   /* now clip against each of the planes that might cut the polygon,
    * at each step toggling between polygons p1 and p2
    */
   p = p1;
   q = &p2;
   if (x0out) CLIP_AND_SWAP(sx, -1., box->x0, p, q, r);
   if (x1out) CLIP_AND_SWAP(sx,  1., box->x1, p, q, r);
   if (y0out) CLIP_AND_SWAP(sy, -1., box->y0, p, q, r);
   if (y1out) CLIP_AND_SWAP(sy,  1., box->y1, p, q, r);
   if (z0out) CLIP_AND_SWAP(sz, -1., box->z0, p, q, r);
   if (z1out) CLIP_AND_SWAP(sz,  1., box->z1, p, q, r);
   /* if result ended up in p2 then copy it to p1 */
   if (p == &p2)
      memcpy(p1, &p2, sizeof(Poly) - (POLY_NMAX-p2.n)*sizeof(Poly_vert));
   return POLY_CLIP_PARTIAL;
}

/*
 * poly_clip_to_halfspace: clip convex polygon p against a plane,
 * copying the portion satisfying sign*s[index] < k*sw into q,
 * where s is a Poly_vert* cast as a DBL*.
 * index is an index into the array of DBLs at each vertex, such that
 * s[index] is sx, sy, or sz (screen space x, y, or z).
 * Thus, to clip against xmin, use
 *      poly_clip_to_halfspace(p, q, XINDEX, -1., -xmin);
 * and to clip against xmax, use
 *      poly_clip_to_halfspace(p, q, XINDEX,  1.,  xmax);
 */

void
poly_clip_to_halfspace(Poly *p, Poly *q, int index, DBL sign, DBL k)
{
   int i, m;
   DBL *up, *vp, *wp;
   Poly_vert *u, *v;
   DBL t, tu, tv;

   q->n = 0;
   q->mask = p->mask;
   /* start with u=vert[n-1], v=vert[0] */
   u = &p->vert[p->n-1];
   tu = sign*COORD(u, index) - u->sw * k;
   for (v= &p->vert[0], i=p->n; i>0; i--, u=v, tu=tv, v++) {
      /* on old polygon (p), u is previous vertex,v is the
       * current vertex tv is negative if vertex v is in */
      tv = sign*COORD(v, index) - v->sw*k;
      if (tu<=0. ^ tv<=0.) {
         /* edge crosses plane; add intersection point to q */
         t = tu / (tu-tv);
         up = (DBL *)u;
         vp = (DBL *)v;
         wp = (DBL *)&q->vert[q->n];
         for (m=p->mask; m!=0; m>>=1, up++, vp++, wp++)
            if (m&1) *wp = *up + t*(*vp - *up);
         q->n++;
         }
      if (tv<=0.)  /* vertex v is in, copy it to q */
         q->vert[q->n++] = *v;
      }
}

/* mx4_transform: transform 4-vector p by matrix m */
/*  yielding q: q = p*m */
static void
mx4_transform(px, py, pz, pw, m, qxp, qyp, qzp, qwp)
   DBL px, py, pz, pw, *qxp, *qyp, *qzp, *qwp;
   MATRIX m;
{
   *qxp = px*m[0][0] + py*m[1][0] + pz*m[2][0] + pw*m[3][0];
   *qyp = px*m[0][1] + py*m[1][1] + pz*m[2][1] + pw*m[3][1];
   *qzp = px*m[0][2] + py*m[1][2] + pz*m[2][2] + pw*m[3][2];
   *qwp = px*m[0][3] + py*m[1][3] + pz*m[2][3] + pw*m[3][3];
}

/* Using the three vectors:
      P  - Point in object space to shade,
      I  - Incident vector (direction from eye to P),
      N  - Normal to the surface at P
   Determine the amount of contribution of the ambient, diffuse,
   and specular components of the light.
   
   The lighting model used here is the standard Phong model.
*/
static void
shade_point(VECTOR P, VECTOR N, VECTOR LP, DBL phong_coeff, DBL *d, DBL *s) 
{
   VECTOR I, L, H;
   DBL dot;

   VecSub(Eye->Orientation.Origin, P, I);
   if (VecNormalize(&I) < 0.0) {
      printf("Incident vector is null\n");
      exit(1);
      }
   VecSub(LP, P, L);
   if (VecNormalize(&L) < 0.0) {
      printf("Light vector is null\n");
      exit(1);
      }
   dot = VecDot(L, N);
/* printf("L: <%f, %f, %f> len = %f, dot = %f\n",
       L.x, L.y, L.z, VecLen(L), dot); */
   if (dot < EPSILON) {
      /* The normal points away from the light, no shading here */
      /* printf("Normal points away\n"); */
      *d = 0.0;
      *s = 0.0;
      }
   else {
      /* Diffuse contributions from light source */
      *d = dot;
      /* Specular contributions from light source */
      VecH(I, L, &H);
      dot = VecDot(N, H);
      if (dot > 0.0)
         *s = pow(dot, phong_coeff);
      else
         *s = 0.0;
      }
}

static void
Determine_Color(pos, nor, Color)
   VECTOR pos, nor;
   VECTOR *Color;
{
   VECTOR Surface_Color;
   OBJECT *Light_Source;
   DBL d, s;

   /* Do a basic lighting model */
   Surface_Color.x = Current_Display_Object->Texture.Color[0];
   Surface_Color.y = Current_Display_Object->Texture.Color[1];
   Surface_Color.z = Current_Display_Object->Texture.Color[2];

/*
printf("Determine color at (%f, %f, %f) for (%f, %f, %f)\n",
       pos.x, pos.y, pos.z,
       Surface_Color.x, Surface_Color.y, Surface_Color.z);
*/

   /* Ambient lighting */
   Color->x = Surface_Color.x * Global_Ambient;
   Color->y = Surface_Color.y * Global_Ambient;
   Color->z = Surface_Color.z * Global_Ambient;

   /* Diffuse and Specular lighting */
   for (Light_Source = Light_Sources ; 
        Light_Source != NULL;
        Light_Source = Light_Source->Next_Object) {

/* printf("Check light %p\n", Light_Source); */

      shade_point(pos, nor, Light_Source->Orientation.Origin,
                  Phong_Coeff, &d, &s);

      /* Diffuse contribution */
      Color->x += d * Surface_Color.x * Global_Diffuse;
      Color->y += d * Surface_Color.y * Global_Diffuse;
      Color->z += d * Surface_Color.z * Global_Diffuse;

      /* Specular contribution */
      Color->x += s;
      Color->y += s;
      Color->z += s;
/* printf("Diffuse: %f, Specular: %f\n", d, s); */
      }
}

/* called at each pixel by poly_scan */
static void
pixelproc(x, y, pt)     
   int x, y;
   Poly_vert *pt;
{
   char r, g, b;
   unsigned int sz;
   VECTOR nor, pos, tvec;
   DBL len;
   VECTOR Color;
/* printf("Pixelproc(%d, %d)\n", x, y); */
   MakeVec(pt->x, pt->y, pt->z, pos);
   VecSub(Eye->Orientation.Origin, pos, tvec);
   /* Display the point on the screen */
   y = YResolution/2 - y;
   x = x + XResolution/2;
   sz = (unsigned int)(10.0 * sqrt(VecDot(tvec, tvec)));
   if (sz < get_depth(x, y)) {
      /* unitize the normal vector */
      MakeVec(pt->nx, pt->ny, pt->nz, nor);
      len = VecDot(nor, nor);
      if (fabs(len) < EPSILON) {
         printf("Bad normal on triangle: <%f, %f, %f> -> %f\n",
                pt->nx, pt->ny, pt->nz, len);
         exit(1);
         }
      VInverseScale(nor, nor, len);

      /* Find out what the color is for this point */
      Determine_Color(pos, nor, &Color);
      if (Color.x > 1.0) Color.x = 1.0;
      if (Color.y > 1.0) Color.y = 1.0;
      if (Color.z > 1.0) Color.z = 1.0;

      if (x == XResolution || y == YResolution)
         return;
      else if (x < 0 || x > XResolution || y < 0 || y > YResolution) {
         printf("Bad coordinate (%d, %d)\n", x, y);
         exit(1);
         }

/* printf("Color = (%f, %f, %f)\n", Color.x, Color.y, Color.z); */

      putpixel(x, y, get_indexed_color(Color));

      /* Save the current depth */
      put_depth(x, y, sz);
      put_color(x, y, Color.x, Color.y, Color.z);
      }
}

void
scan_convert(p, WS)
   Poly *p;
   TRANSFORMATION *WS;
{
   int i, r, y;
   Poly_vert *v;

   /* transform vertices from world space to homogeneous screen 
      space */
   for (i=0; i<p->n; i++) {
      v = &p->vert[i];
      mx4_transform(v->x, v->y, v->z, 1.0, WS->matrix,
                    &v->sx, &v->sy, &v->sz, &v->sw);
      }

   /* interpolate sx, sy, sz, sw, nx, ny, nz in poly_clip */
   p->mask = POLY_MASK(sx) | POLY_MASK(sy) | POLY_MASK(sz) | 
             POLY_MASK(sw) | POLY_MASK(nx) | POLY_MASK(ny) | 
             POLY_MASK(nz);
   /* poly_print("before clipping", p); */
   if ((r = poly_clip_to_box(p, &box)) == POLY_CLIP_OUT) {
      /* If totally clipped polygon then quit*/
      /* printf("Poly totally clipped\n"); */
      return;
      }
   else if (r == POLY_CLIP_PARTIAL) {
      /* poly_print("After clipping", p);
      printf("Poly partially clipped\n"); */
      }
   else {
      /* printf("Poly not clipped\n"); */
      }

   /* do homogeneous division of screen position, texture position */
   for (i=0; i<p->n; i++) {
      v = &p->vert[i];
      v->sx /= v->sw;
      v->sy /= v->sw;
      v->sz /= v->sw;
      }

   /* See if the polygon is within the current strip */
   for (i=0; i<p->n; i++) {
      y = YResolution/2 - p->vert[i].sy;
      if (y >= First_Map_Line && y <= Last_Map_Line)
         goto out;
      }
   return;
   out:

   win.y0 = YResolution/2 - Last_Map_Line;
   win.y1 = YResolution/2 - First_Map_Line;

   /* previously we ignored q (assumed q=1),
    * henceforth ignore sw (assume sw=1)
    * Interpolate sx, sy, sz, nx, ny, nz in poly_scan
    */
   p->mask |= POLY_MASK(x) | POLY_MASK(y) | POLY_MASK(z);
   p->mask &= ~POLY_MASK(sw);
   /* poly_print("scan converting:", p); */
   poly_scan(p, &win, pixelproc);       /* scan convert! */
}

/*
 * poly_scan: Scan convert a polygon, calling pixelproc at 
 * each pixel with an interpolated Poly_vert structure.
 * Polygon can be clockwise or ccw.
 * Polygon is clipped in 2-D to win, the screen space window.
 *
 * Scan conversion is done on the basis of Poly_vert fields sx and sy.
 * These two must always be interpolated, and only they have 
 * special meaning to this code; any other fields are 
 * blindly interpolated regardless of their semantics.
 *
 * The pixelproc subroutine takes the arguments:
 *
 *      pixelproc(x, y, point)
 *      int x, y;
 *      Poly_vert *point;
 *
 * All the fields of point indicated by p->mask will be 
 * valid inside pixelproc except sx and sy.
 * If they were computed, they would have values
 * sx=x+.5 and sy=y+.5, since sampling is done at pixel centers.
 */

void
poly_scan(p, win, pixelproc)
   Poly *p;     /* polygon */
   Window *win;         /* 2-D screen space clipping window */
   void (*pixelproc)(); /* procedure called at each pixel */
{
   int i, li, ri, y, ly, ry, top, rem, mask;
   DBL ymin;
   Poly_vert l, r, dl, dr;

   if (p->n>POLY_NMAX) {
      fprintf(stderr, "poly_scan: too many vertices: %d\n", p->n);
      return;
      }
   ymin = HUGE;
   for (i=0; i<p->n; i++)       /* find top vertex (y points down) */
      if (p->vert[i].sy < ymin) {
         ymin = p->vert[i].sy;
         top = i;
         }

   li = ri = top;               /* left and right vertex indices */
   rem = p->n;                  /* number of vertices remaining */
   y = (int)ceil(ymin-.5);   /* current scan line */
   ly = ry = y-1;               /* lower end of left & right edges */
   mask = p->mask & ~POLY_MASK(sy);/* stop interpolating screen y */
   while (rem > 0) {
      /* scan in y, activating new edges on left & right as scan
         line passes over new vertices */
      while (ly<=y && rem>0) {  /* advance left edge? */
         rem--;
         i = li-1;                      /* step ccw down left side */
         if (i<0) i = p->n-1;
         incrementalize_y((DBL *)&p->vert[li], (DBL *)&p->vert[i], (DBL *)&l,
                          (DBL *)&dl, y, mask);
         ly = (int)floor(p->vert[i].sy+.5);
         li = i;
         }
      while (ry<=y && rem>0) {  /* advance right edge? */
         rem--;
         i = ri+1;                      /* step cw down right edge */
         if (i>=p->n) i = 0;
         incrementalize_y((DBL *)&p->vert[ri], (DBL *)&p->vert[i], (DBL *)&r,
                          (DBL *)&dr, y, mask);
         ry = (int)floor(p->vert[i].sy+.5);
         ri = i;
         }
      while (y<ly && y<ry) {
         /* do scanlines till end of l or r edge */
         if (y>=win->y0 && y<=win->y1) {
            if (l.sx<=r.sx)
               scanline(y, &l, &r, win, pixelproc, mask);
            else
               scanline(y, &r, &l, win, pixelproc, mask);
            }
         y++;
         increment((DBL *)&l, (DBL *)&dl, mask);
         increment((DBL *)&r, (DBL *)&dr, mask);
         }
      }
}

/* scanline: output scanline by sampling polygon at Y=y+.5 */

static void
scanline(y, l, r, win, pixelproc, mask)
   int y, mask;
   Poly_vert *l, *r;
   Window *win;
   void (*pixelproc)(int, int, Poly_vert *);
{
   int x, lx, rx;
   Poly_vert p, dp;
   mask &= ~POLY_MASK(sx);              /* stop interpolating screen x */
   lx = (int)ceil(l->sx-.5);
   if (lx<win->x0) lx = win->x0;
   rx = (int)floor(r->sx-.5);
   if (rx>win->x1) rx = win->x1;
   if (lx>rx) return;
   incrementalize_x((DBL *)l, (DBL *)r, (DBL *)&p, (DBL *)&dp, lx, mask);
   for (x=lx; x<=rx; x++) {             /* scan in x, generating pixels */
      (*pixelproc)(x, y, &p);
      increment((DBL *)&p, (DBL *)&dp, mask);
      }
}

/*
 * incrementalize_y: put intersection of line Y=y+.5 with edge  
 * between points p1 and p2 in p, put change with respect to y in dp  
 */

static void
incrementalize_y(p1, p2, p, dp, y, mask)
   DBL *p1, *p2, *p, *dp;
   int mask, y;
{
   DBL dy, frac;
   dy = ((Poly_vert *)p2)->sy - ((Poly_vert *)p1)->sy;
   if (dy==0.0) dy = 1.0;
   frac = y+.5 - ((Poly_vert *)p1)->sy;
   for (; mask!=0; mask>>=1, p1++, p2++, p++, dp++)
      if (mask&1) {
         *dp = (*p2 - *p1) / dy;
         *p = *p1 + *dp * frac;
         }
}


/*
 * incrementalize_x: put intersection of line X=x+.5 with 
 * edge between points p1 and p2 in p,
 *  put change with respect to x in dp
 */

static void
incrementalize_x(p1, p2, p, dp, x, mask)
   register DBL *p1, *p2, *p, *dp;
   register int mask;
   int x;
{
   DBL dx, frac;
   dx = ((Poly_vert *)p2)->sx - ((Poly_vert *)p1)->sx;
   if (dx==0.) dx = 1.;
   frac = x+.5 - ((Poly_vert *)p1)->sx;
   for (; mask!=0; mask>>=1, p1++, p2++, p++, dp++)
      if (mask&1) {
         *dp = (*p2 - *p1) / dx;
         *p = *p1 + *dp * frac;
         }
}

static void
increment(p, dp, mask)
   register DBL *p, *dp;
   register int mask;
{
   for (; mask!=0; mask>>=1, p++, dp++)
      if (mask & 1)
         *p += *dp;
}

void
render(poly, eval, WS)
   Poly *poly;
   void (*eval)(Poly *, DBL, DBL, Poly_vert *);
   TRANSFORMATION *WS;
{
   int i, j;
   DBL u, v;
   DBL u_min = 0.0;
   DBL u_max = 1.0;
   DBL v_min = 0.0;
   DBL v_max = 1.0;
   int u_steps = (int)poly->vert[0].x;
   int v_steps = (int)poly->vert[0].y;
   DBL delta_u, delta_v;
   Poly Polygon;

   Polygon.n = 4;
   delta_u = (u_max - u_min) / (DBL)u_steps;
   delta_v = (v_max - v_min) / (DBL)v_steps;

   /* Dump out polygons */
   for (i=0,u=u_min;i<u_steps;i++,u+=delta_u) {
      for (j=0,v=v_min;j<v_steps;j++,v+=delta_v) {
         eval(poly, u, v, &Polygon.vert[0]);
         eval(poly, u, v+delta_v, &Polygon.vert[1]);
         eval(poly, u+delta_u, v+delta_v, &Polygon.vert[2]);
         eval(poly, u+delta_u, v, &Polygon.vert[3]);
         scan_convert(&Polygon, WS);
         }
      }
}

void
setup_renderer()
{
   /* 3-D screen clipping window */
   box.x0 = -(DBL)XResolution/2.0;
   box.x1 =  (DBL)XResolution/2.0;
   box.y0 = -(DBL)YResolution/2.0;
   box.y1 =  (DBL)YResolution/2.0;
   box.z0 =  -9.0;
   box.z1 =  32678.0;

   /* 2-D screen clipping window */
   win.x0 = -XResolution/2;
   win.x1 =  XResolution/2;
   win.y0 = -YResolution/2;
   win.y1 =  YResolution/2;
}
